/*******************************************************************************
 * $Id: network.c 206 2008-08-01 21:44:50Z msiegen $
 *
 * Thermostat Network Interface
 *
 * by Michael Siegenthaler
 * http://www.msigi.net/
 *
 * THIS SOFTWARE IS PUBLIC DOMAIN AND IS PROVIDED WITHOUT ANY WARRANTY
 *
 ******************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "config.h" // must come before other includes
#include "network.h"

/******************************************************************************/

int device_sd = -1;


// Connect to the device serial port
// Return 0 on success, -1 on failure
int initialize_port_net() {

  int rv, i, delay_usec;
  double random;

  struct hostent *he;
  struct sockaddr_in inst_ip_addr;

  // Initialize the random number generator
  srand((unsigned int)getpid());

  // Look up hostname
  if ((he=gethostbyname( DEVSERVER_IP ))==NULL) {
    return -1;
  }

  for (i=0; i<7; i++) {

    // Create socket
    if ((device_sd = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
      return -1;
    }

    // Set up address structure
    inst_ip_addr.sin_family = AF_INET;
    inst_ip_addr.sin_port = htons( DEVSERVER_PORT );
    inst_ip_addr.sin_addr = *((struct in_addr *)he->h_addr);
    memset(&(inst_ip_addr.sin_zero), '\0', 8);

    // Connect
    rv = connect(device_sd, (struct sockaddr *)&inst_ip_addr, sizeof(struct sockaddr));
    if (rv == -1) {
      if (errno == ECONNREFUSED) {
        // The connection was refused. Probably somebody else is using the
        // device server, so we should linearly back off and retry.
        random = (double)rand() / (double)RAND_MAX;
        // Delay for at least half a second, and some random amount longer
        delay_usec = 500000 + (int)(1000000 * (i+1) * random);
        //printf("Connection refused %f, %d\n", random, delay_usec);
        usleep(delay_usec);
        continue;
      } else {
        close(device_sd);
        //perror("other error");
        // An error occurred that was not a connection refusal. Don't bother retrying.
        return -1;
      }
    }

    return 0;

  }

  return -1;
}

// Send a command to the thermostat
// Return 0 on success, -1 on failure
int send_command_net(char* buf, int length) {

  int bytes_written;
  int bytes_already_written = 0;

  // Make sure port has been opened
  if (device_sd < 0) return -1;

  // Flush the I/O buffers, in case a previous command still has reply data waiting
  // Note: This doesn't seem to work reliably because the device server does not
  // transmit bufferred bytes immediately on connection (or maybe the network latency
  // is the issue). Still, the flush may be needed if a large amount of data is stuck
  // in the device server's buffer.
  flush_net_input();

  // Write the buffer out to the serial port in a loop
  while (bytes_already_written < length) {
    // Write to port
    //printf("Writing %02x %02x %02x %02x %02x\n", buf[0] & 0xff, buf[1] & 0xff, buf[2] & 0xff, buf[3] & 0xff, buf[4] & 0xff);
    bytes_written = send(device_sd, buf+bytes_already_written, length-bytes_already_written, 0);
    if (bytes_written < 0 )
      return -1;
    bytes_already_written += bytes_written;
  }

  return 0;
}

// Receive a reply from the thermostat
// Return length of received reply, or -1 on error
int receive_reply_net(char* buf, int maxlen) {

  int i;
  int bytes_read;
  int bytes_already_read;
  struct timeval tv;
  fd_set readfds;
  int msg_length = 0; // refers to data section only
  int cksum;

  // Make sure port has been opened
  if (device_sd < 0) return -1;

  // Make sure the buffer is long enough for the shortest possible message
  if (maxlen < 3) return -1;

  bytes_already_read = 0;

  // set timeout for read from socket
  tv.tv_sec = 10; // set read timeout to 10 seconds
  tv.tv_usec = 0;

  while (1) {

    FD_ZERO(&readfds);
    FD_SET(device_sd, &readfds);

    // Check if data is available on serial port
    if ( select(device_sd+1, &readfds, NULL, NULL, &tv) <= 0 ) {
      // Timed out, or error
      return -1;
    }

    if (FD_ISSET(device_sd, &readfds)) { // we have input

      // Read from serial port
      bytes_read = recv(device_sd, buf+bytes_already_read, maxlen-bytes_already_read, 0);

      //printf("read %d bytes from serial port: ", bytes_read);
      //for (i=0; i<bytes_read; i++) {
      //  printf("0x%02x ", *(unsigned char*)(buf+bytes_already_read+i));
      //}
      //printf("\n");

      if (bytes_read < 0)
        return -1;

      bytes_already_read += bytes_read;

      if (bytes_already_read >= 2) {
        // Update the message length
        msg_length = (buf[1] >> 4) & 0x0f;
      }

      if (bytes_already_read >= msg_length+3) {
        // Complete message received...verify the checksum
        cksum = 0;
        for (i=0; i<msg_length+2; i++)
          cksum += (buf[i] & 0xff);
        if ((cksum & 0xff) != (buf[msg_length+2] & 0xff)) {
          //printf("Checksum does not match: should be %02x\n", cksum & 0xff);
          return -1; // checksum mismatch!
        } else {
          // Good!
          return bytes_already_read;
        }
      }
    }
  }

  return -1;
}


#define FLUSH_BUF_LEN 1500

// Flush the input buffer for the socket
// Return 0 on success, or -1 on error
void flush_net_input() {

  int bytes_read;
  struct timeval tv;
  fd_set readfds;
  char buf[FLUSH_BUF_LEN];

  // Make sure port has been opened
  if (device_sd < 0) return;

  // Set a zero timeout because we don't want to wait for data
  tv.tv_sec = 0;
  tv.tv_usec = 1;

  while (1) {

    FD_ZERO(&readfds);
    FD_SET(device_sd, &readfds);

    // Check if data is available on serial port
    if ( select(device_sd+1, &readfds, NULL, NULL, &tv) < 0 ) {
      // Error, let's return and hope for the best...
      return;
    }

    if (FD_ISSET(device_sd, &readfds)) { // we have some data to flush

      // Read from serial port
      bytes_read = recv(device_sd, buf, FLUSH_BUF_LEN, 0);

      if (bytes_read < 0)
        return; // must be an error of some sort

    } else { // we don't have anything to flush

      return;

    }
  }

}
